Unit TI92;

{****************************************************************************}
Interface
{****************************************************************************}

Uses List,TIdef;

Const
  c_wait    = $09;
  c_OK      = $56;
  c_request = $68;
  c_screen_dump = $6D;
  c_remote  = $87;
  c_eot     = $92;

Procedure Get_avail_ports;
Procedure TIreset_lpt;
Function  TIget_lpt : Byte;
Procedure TIsend_lpt (x : Byte);
Procedure TIreset_com;
Function  TIget_com : Byte;
Procedure TIsend_com (x : Byte);
Function  TIget_command : Byte;
Procedure TIsend_command (x,a,b : Byte);
Procedure TIremote (key : Word);
Procedure TIremoteStr (s : String);
Procedure TIget_packet (pack : ppacket; bar : Boolean; check : Boolean);
Procedure TIsend_packet (pack : ppacket; bar : Boolean);
Procedure TIget_dir (Var l : lst);
Function  TIget_screen (Var buffer : Pointer) : Word;
Procedure TIget_var (x : pTIvar);
Procedure TIsend_var (x : pTIvar);
Function  TIget_backup (Var p : mpack; Var ROM : TIstr) : Word;
Procedure TIsend_backup (fname : String);

Const
  TIreset : Procedure = TIreset_lpt;
  TIget   : Function : Byte = TIget_lpt;
  TIsend  : Procedure (x : Byte) = TIsend_lpt;

{****************************************************************************}
Implementation
{****************************************************************************}

Uses Debug,Timer,Txt;

Const
  dir_req : Array [0..5] Of Byte = ($00,$00,$00,$00,$19,$00);
  p_dir_request : packet = (typ:$A2; size:6; data:@dir_req);
  bckp_req : Array [0..16] Of Byte
           = ($00,$00,$00,$00,$1D,$0B,$6D,$61,
              $69,$6E,$5C,$62,$61,$63,$6B,$75,
              $70);
  p_bckp_request : packet = (typ:$A2; size:17; data:@bckp_req);

Procedure Get_avail_ports;
Var
  a : Byte;
  w : Word;
Begin
  w :=0;
  For a :=1 To 8 Do
  Begin
    ports_add[a] :=MemW[$40:w];
    Inc (w,2);
  End;
End;

Procedure TIreset_lpt; Assembler;
Asm
  mov  dx,P_out
  mov  al,3
  mov  di,P_iteration
  @poll:
       out  dx,al
       dec  di
  jnz  @poll
End;

Function TIget_lpt : Byte; Assembler;
Asm
  push es
  mov  ax,40h
  mov  es,ax
  mov  TIerror,0
  xor  bl,bl
  mov  cl,8

  @main:
       mov  dx,P_in
       mov  ch,es:[6Ch]
       @wait1:
            {in polling 1}
            mov  di,P_iteration
            @in_poll1:
                 in   al,dx
                 dec  di
            jnz  @in_poll1
            and  al,30h
            cmp  al,30h
            jne  @ok1
            mov  di,es:[1Ah]
            cmp  di,es:[1Ch]
            je   @time1
            mov  al,es:[di]
            cmp  al,1Bh
            jne  @time1
                 mov  TIerror,1
                 jmp  @quit
            @time1:
            mov  al,es:[6Ch]
            sub  al,ch
            cmp  al,TItime_out
            jb   @wait1
                 mov  TIerror,2
                 jmp  @quit
       @ok1:
       shr  bl,1
       cmp  al,10h
       jne  @bit0

       @bit1:
            or   bl,80h
            mov  dx,P_out
            mov  al,1
            {out polling 1}
            mov  di,P_iteration
            @out_poll1:
                 out  dx,al
                 dec  di
            jnz  @out_poll1
            mov  dx,P_in
            mov  ch,es:[6Ch]
            @wait2:
                 {in polling 2}
                 mov  di,P_iteration
                 @in_poll2:
                      in   al,dx
                      dec  di
                 jnz  @in_poll2
                 and  al,20h
                 jnz  @next
                 mov  di,es:[1Ah]
                 cmp  di,es:[1Ch]
                 je   @time2
                 mov  al,es:[di]
                 cmp  al,1Bh
                 jne  @time2
                      mov  TIerror,1
                      jmp  @quit
                 @time2:
                 mov  al,es:[6Ch]
                 sub  al,ch
                 cmp  al,TItime_out
                 jb   @wait2
                      mov  TIerror,2
                      jmp  @quit

       @bit0:
            mov  dx,P_out
            mov  al,2
            {out polling 2}
            mov  di,P_iteration
            @out_poll2:
                 out  dx,al
                 dec  di
            jnz  @out_poll2
            mov  dx,P_in
            mov  ch,es:[6Ch]
            @wait3:
                 mov  di,P_iteration
                 {in polling 3}
                 @in_poll3:
                      in   al,dx
                      dec  di
                 jnz  @in_poll3
                 and  al,10h
                 jnz   @next
                 mov  di,es:[1Ah]
                 cmp  di,es:[1Ch]
                 je   @time3
                 mov  al,es:[di]
                 cmp  al,1Bh
                 jne  @time3
                      mov  TIerror,1
                      jmp  @quit
                 @time3:
                 mov  al,es:[6Ch]
                 sub  al,ch
                 cmp  al,TItime_out
                 jb   @wait3
                      mov  TIerror,2
                      jmp  @quit

       @next:
            mov  dx,P_out
            mov  al,3
            {out polling 3}
            mov  di,P_iteration
            @out_poll3:
                 out  dx,al
                 dec  di
            jnz  @out_poll3
            dec  cl
            jnz  @main

  mov  al,bl
  @quit:
  pop  es
End;

Procedure TIsend_lpt (x : Byte); Assembler;
Asm
  push es
  mov  ax,40h
  mov  es,ax
  mov  TIerror,0
  mov  bl,x
  mov  cl,8

  @main:
       mov  dx,P_out
       mov  al,3
       {out polling 1}
       mov  di,P_iteration
       @out_poll1:
            out  dx,al
            dec  di
       jnz  @out_poll1

       test bl,1
       jz   @bit0

       @bit1:
            mov  dx,P_out
            mov  al,2
            {out polling 2}
            mov  di,P_iteration
            @out_poll2:
                 out  dx,al
                 dec  di
            jnz  @out_poll2
            mov  dx,P_in
            mov  ch,es:[6Ch]
            @wait1:
                 {in polling 1}
                 mov  di,P_iteration
                 @in_poll1:
                      in   al,dx
                      dec  di
                 jnz  @in_poll1
                 and  al,10h
                 jz  @ok1
                 mov  di,es:[1Ah]
                 cmp  di,es:[1Ch]
                 je   @time1
                 mov  al,es:[di]
                 cmp  al,1Bh
                 jne  @time1
                      mov  TIerror,1
                      jmp  @quit
                 @time1:
                 mov  al,es:[6Ch]
                 sub  al,ch
                 cmp  al,TItime_out
                 jb   @wait1
                      mov  TIerror,2
                      jmp  @quit
            @ok1:
            mov  dx,P_out
            mov  al,3
            {out polling 3}
            mov  di,P_iteration
            @out_poll3:
                 out  dx,al
                 dec  di
            jnz  @out_poll3
            mov  dx,P_in
            mov  ch,es:[6Ch]
            @wait2:
                 {in polling 2}
                 mov  di,P_iteration
                 @in_poll2:
                      in   al,dx
                      dec  di
                 jnz  @in_poll2
                 and  al,10h
                 jnz  @next
                 mov  di,es:[1Ah]
                 cmp  di,es:[1Ch]
                 je   @time2
                 mov  al,es:[di]
                 cmp  al,1Bh
                 jne  @time2
                      mov  TIerror,1
                      jmp  @quit
                 @time2:
                 mov  al,es:[6Ch]
                 sub  al,ch
                 cmp  al,TItime_out
                 jb   @wait2
                      mov  TIerror,2
                      jmp  @quit

       @bit0:
            mov  dx,P_out
            mov  al,1
            {out polling 4}
            mov  di,P_iteration
            @out_poll4:
                 out  dx,al
                 dec  di
            jnz  @out_poll4
            mov  dx,P_in
            mov  ch,es:[6Ch]
            @wait3:
                 {in polling 3}
                 mov  di,P_iteration
                 @in_poll3:
                      in   al,dx
                      dec  di
                 jnz  @in_poll3
                 and  al,20h
                 jz   @ok3
                 mov  di,es:[1Ah]
                 cmp  di,es:[1Ch]
                 je   @time3
                 mov  al,es:[di]
                 cmp  al,1Bh
                 jne  @time3
                      mov  TIerror,1
                      jmp  @quit
                 @time3:
                 mov  al,es:[6Ch]
                 sub  al,ch
                 cmp  al,TItime_out
                 jb   @wait3
                      mov  TIerror,2
                      jmp  @quit
            @ok3:
            mov  dx,P_out
            mov  al,3
            {out polling 5}
            mov  di,P_iteration
            @out_poll5:
                 out  dx,al
                 dec  di
            jnz  @out_poll5
            mov  dx,P_in
            mov  ch,es:[6Ch]
            @wait4:
                 {in polling 4}
                 mov  di,P_iteration
                 @in_poll4:
                      in   al,dx
                      dec  di
                 jnz  @in_poll4
                 and  al,20h
                 jnz  @next
                 mov  di,es:[1Ah]
                 cmp  di,es:[1Ch]
                 je   @time4
                 mov  al,es:[di]
                 cmp  al,1Bh
                 jne  @time4
                      mov  TIerror,1
                      jmp  @quit
                 @time4:
                 mov  al,es:[6Ch]
                 sub  al,ch
                 cmp  al,TItime_out
                 jb   @wait4
                      mov  TIerror,2
                      jmp  @quit

       @next:
            shr  bl,1
            dec  cl
            jnz  @main
  @quit:
  pop  es
End;

Procedure TIreset_com; Assembler;
Asm
  mov  dx,S_base
  add  dx,3
  mov  al,83h
  out  dx,al     {LCR: 8 bits/ no parity/ 1 stop bit + DLAB=1}
  jmp  @next1
  @next1:

  mov  dx,S_base
  mov  al,0Ch
  out  dx,al     {divisor LSB}
  jmp  @next2
  @next2:
  inc  dx
  mov  al,00h
  out  dx,al     {divisor MSB}
  jmp  @next3
  @next3:

  mov  dx,S_base
  add  dx,3
  mov  al,03h
  out  dx,al      {LCR: DLAB=0}
  jmp  @next4
  @next4:

  inc  dx
  mov  al,0
  out  dx,al      {no interrupt}
  jmp  @next5
  @next5:
  inc  dx
  out  dx,al      {16550 FIFO disabled}
  jmp  @next6
  @next6:

  mov  dx,S_base
  add  dx,4
  mov  al,03h
  out  dx,al      {activate RTS & DTR}
  jmp  @next7
  @next7:
End;

Function  TIget_com : Byte; Assembler;
Asm
  mov  ax,40h
  mov  es,ax
  mov  TIerror,0

  {While (Port[COM+5] And 96)<96 Do;}
  mov  dx,S_base
  add  dx,5
  mov  ch,es:[6Ch]
  @wait1:
       in   al,dx
       and  al,60h
       cmp  al,60h
       je   @next
       mov  di,es:[1Ah]
       cmp  di,es:[1Ch]
       je   @nokey1
       mov  al,es:[di]
       cmp  al,1Bh
       jne  @nokey1
            mov  TIerror,1
            jmp  @quit
       @nokey1:
       mov  al,es:[6Ch]
       sub  al,ch
       cmp  al,TItime_out
       jb   @wait1
            mov  TIerror,2
            jmp  @quit
  {read Port[COM];}
  @next:
  mov  dx,S_base
  in   al,dx
  {While (Port[COM+5] And 1)=0 Do;}
  add  dx,5
  mov  ch,es:[6Ch]
  @wait2:
       in   al,dx
       and  al,1
       jnz  @get
       mov  di,es:[1Ah]
       cmp  di,es:[1Ch]
       je   @nokey2
       mov  al,es:[di]
       cmp  al,1Bh
       jne  @nokey2
            mov  TIerror,1
            jmp  @quit
       @nokey2:
       mov  al,es:[6Ch]
       sub  al,ch
       cmp  al,TItime_out
       jb   @wait2
            mov  TIerror,2
            jmp  @quit
  {Get :=Port[COM];}
  @get:
  mov  dx,S_base
  in   al,dx

  @quit:
End;

Procedure TIsend_com (x : Byte); Assembler;
Asm
  mov  ax,40h
  mov  es,ax
  mov  TIerror,0

  {read Port[COM]}
  mov  dx,S_base
  in   al,dx
  {While (Port[COM+5] And 1)=1 Do;}
  add  dx,5
  mov  ch,es:[6Ch]
  @wait1:
       in   al,dx
       and  al,1
       jz   @next
       mov  di,es:[1Ah]
       cmp  di,es:[1Ch]
       je   @nokey1
       mov  al,es:[di]
       cmp  al,1Bh
       jne  @nokey1
            mov  TIerror,1
            jmp  @quit
       @nokey1:
       mov  al,es:[6Ch]
       sub  al,ch
       cmp  al,TItime_out
       jb   @wait1
            mov  TIerror,2
            jmp  @quit
  {While (Port[COM+5] And 96)<96 Do;}
  @next:
  mov  ch,es:[6Ch]
  @wait2:
       in   al,dx
       and  al,60h
       cmp  al,60h
       je   @send
       mov  di,es:[1Ah]
       cmp  di,es:[1Ch]
       je   @nokey2
       mov  al,es:[di]
       cmp  al,1Bh
       jne  @nokey2
            mov  TIerror,1
            jmp  @quit
       @nokey2:
       mov  al,es:[6Ch]
       sub  al,ch
       cmp  al,TItime_out
       jb   @wait2
            mov  TIerror,2
            jmp  @quit
  {Port[COM] :=x;}
  @send:
  mov  dx,S_base
  mov  al,x
  out  dx,al

  @quit:
End;

Function TIget_command : Byte;
Var
  x : Byte;
Label
  failure;
Begin
  If (TIget<>$89) And (TIerror=0) Then
    TIerror :=5;
  If TIerror>0 Then Goto failure;
  x :=TIget;
  If TIerror>0 Then Goto failure;
  TIget_command :=x;
  TIget;
  If TIerror>0 Then Goto failure;
  If (TIget>0) And (TIerror=0) And (x=$56) Then
    TIerror :=3;
  failure:
End;

Procedure TIsend_command (x,a,b : Byte);
Label
  failure;
Begin
  TIsend ($09);
  If TIerror>0 Then Goto failure;
  TIsend (x);
  If TIerror>0 Then Goto failure;
  TIsend (a);
  If TIerror>0 Then Goto failure;
  TIsend (b);
  failure:
End;

Procedure TIremote (key : Word);
Label
  failure;
Begin
  TIsend ($09);
  If TIerror>0 Then Goto failure;
  TIsend ($87);
  If TIerror>0 Then Goto failure;
  TIsend (key And $FF);
  If TIerror>0 Then Goto failure;
  TIsend (key Shr 8);
  If TIerror>0 Then Goto failure;
  If (TIget_command<>$56) And (TIerror=0) Then
    TIerror :=5;
  failure:
End;

Procedure TIremoteStr (s : String);
Var
  a : Byte;
Label
  failure;
Begin
  For a :=1 To Length (s) Do
  Begin
    TIremote (Ord (s[a]));
    If TIerror>0 Then Goto failure;
  End;
  failure:
End;

Procedure TIget_packet (pack : ppacket; bar : Boolean; check : Boolean);
Var
  x : Byte;
  a,b,c,o,sum,size,w : Word;
  l : LongInt;
Label
  failure;
Begin
  size :=0;
  If (TIget<>$89) And (TIerror=0) Then
    TIerror :=5;
  If TIerror>0 Then Goto failure;
  x :=TIget;
  If TIerror>0 Then Goto failure;
  pack^.typ :=x;
  w :=TIget;
  If TIerror>0 Then Goto failure;
  Inc (w,TIget Shl 8);
  If TIerror>0 Then Goto failure;
  If w>0 Then
  Begin
    If bar Then
      Text (40,24,#0#127'');
    size :=w;
    GetMem (pack^.data,size); Inc (_malloc,size);
    a :=0;
    l :=0;
    sum :=0;
    For w :=0 To size-1 Do
    Begin
      x :=TIget;
      If TIerror>0 Then Goto failure;
      pack^.data^[w] :=x;
      Inc (sum,x);
      b :=l Div size;
      If bar And (b>a) Then
      Begin
        o :=24*160+80+a Shl 1;
        For c :=a To b Do
        Begin
          Mem[$B800:o] :=219;
          Inc (o,2);
        End;
        a :=b;
      End;
      Inc (l,40);
    End;
    w :=TIget;
    If TIerror>0 Then Goto failure;
    Inc (w,TIget Shl 8);
    If (TIerror=0) And (check And (w<>sum)) Then
      TIerror :=4;
  End;
  pack^.size :=size;

  failure:
  If (TIerror>0) And (size>0) Then Begin
    FreeMem (pack^.data,size); Inc (_mfree,size); End;
  If bar Then
    Text (40,24,'                                        ');
End;

Procedure TIsend_packet (pack : ppacket; bar : Boolean);
Var
  x : Byte;
  a,b,c,sum,size,w,o : Word;
  l : LongInt;
Label
  failure;
Begin
  sum :=0;
  TIsend ($09);
  If TIerror>0 Then Goto failure;
  TIsend (pack^.typ);
  If TIerror>0 Then Goto failure;
  size :=pack^.size;
  TIsend (size And $FF);
  If TIerror>0 Then Goto failure;
  TIsend (size Shr 8);
  If TIerror>0 Then Goto failure;
  If bar Then
    Text (40,24,#0#127'');
  a :=0;
  l :=0;
  For w :=0 To size-1 Do
  Begin
    x :=pack^.data^[w];
    Inc (sum,x);
    TIsend (x);
    If TIerror>0 Then Goto failure;
    b :=l Div size;
    If bar And (b>a) Then
    Begin
      o :=24*160+80+a Shl 1;
      For c :=a To b Do
      Begin
        Mem[$B800:o] :=219;
        Inc (o,2);
      End;
      a :=b;
    End;
    Inc (l,40);
  End;
  TIsend (sum And $FF);
  If TIerror>0 Then Goto failure;
  TIsend (sum Shr 8);
  failure:
  If bar Then
    Text (40,24,'                                        ');
End;

Procedure TIget_dir (Var l : lst);
Var
  x,y,_typ : Byte;
  i,j,w : Word;
  cur_dir,_dir : TIstr;
  pack : packet;
Label
  failure;
Begin
  TIreset;

  If Not skip_TI_reset Then
  Begin
    TIsend_command (c_request,$00,$00);
    If TIerror>0 Then Goto failure;
    If (TIget_command<>$56) And (TIerror=0) Then
      TIerror :=5;
    If TIerror>0 Then Goto failure;
  End;

  TIsend_packet (@p_dir_request,False);
  If TIerror>0 Then Goto failure;
  If (TIget_command<>$56) And (TIerror=0) Then
    TIerror :=5;
  If TIerror>0 Then Goto failure;

  TIget_packet (@pack,False,True);
  If TIerror>0 Then Goto failure;
  y :=pack.data^[5];
  For x :=1 To y Do
    _dir[x] :=Chr (pack.data^[5+x]);
  For x :=y+1 To 8 Do
    _dir[x] :=#32;
  FreeMem (pack.data,pack.size); Inc (_mfree,pack.size);

  cur_dir :=TIblank;
  l.curdir :=0;
  Repeat
    TIsend_command (c_OK,$00,$00);
    If TIerror>0 Then Goto failure;
    TIsend_command (c_wait,$FF,$FF);
    If TIerror>0 Then Goto failure;
    If (TIget_command<>$56) And (TIerror=0) Then
      TIerror :=5;
    If TIerror>0 Then Goto failure;
    TIget_packet (@pack,false,True);
    If TIerror>0 Then Goto failure;

    _typ :=pack.data^[12];
    If _typ=31 Then
      Begin
        Inc (l.curdir);
        i :=0;
        j :=l.curdir;
        New (l.item[l.curdir]); Inc (_malloc,SizeOf (dir));
      End
    Else
      Begin
        Inc (l.item[0]^[l.curdir]^.size);
        i :=l.curdir;
        j :=l.item[0]^[l.curdir]^.size;
      End;
    New (l.item[i]^[j]); Inc (_malloc,SizeOf (TIvar));
    With l.item[i]^[j]^ Do
    Begin
      name :=TIblank;
      For y :=4 To 11 Do
      Begin
        x :=pack.data^[y];
        If x=0 Then
          x :=32;
        name[y-3] :=Chr (x);
      End;
      typ :=_typ;
      sel :=False;
      If _typ=31 Then
        Begin
          size :=1;
          cur_dir :=name;
        End
      Else
        Begin
          dir :=cur_dir;
          size :=pack.data^[14];
          Inc (size,pack.data^[15] Shl 8);
          Inc (size,pack.data^[16] Shl 16);
          Inc (size,pack.data^[17] Shl 24); {var_size+4=data_size}
        End;
    End;
    If _typ=31 Then
    Begin
      New (l.item[j]^[1]); Inc (_malloc,SizeOf (TIvar));
      With l.item[j]^[1]^ Do
      Begin
        name :='..      ';
        typ :=30;
        size :=0;
        sel :=False;
      End;
    End;
    FreeMem (pack.data,pack.size); Inc (_mfree,pack.size);

    TIsend_command (c_OK,$00,$00);
    If TIerror>0 Then Goto failure;
    x :=TIget_command;
    If TIerror>0 Then Goto failure;
  Until x<>$78;
  TIsend_command (c_OK,$00,$00);
  l.item[0]^[0]^.size :=l.curdir;
  l.curdir :=0;

  failure:
End;

Function TIget_screen (Var buffer : Pointer) : Word;
Var
  pack : packet;
Label
  failure;
Begin
  TIget_screen :=0;
  TIreset;
  TIsend_command (c_screen_dump,$00,$00);
  If (TIget_command<>$56) And (TIerror=0) Then
    TIerror :=5;
  If TIerror>0 Then Goto failure;
  TIget_packet (@pack,True,False);
  If (TIerror=0) And (pack.typ<>$15) Then
    TIerror :=5;
  If TIerror>0 Then Goto failure;
  TIget_screen :=pack.size;
  buffer :=pack.data;
  TIsend_command (c_OK,$00,$00);
  failure:
  TIreset;
End;

Procedure TIget_var (x : pTIvar);
Var
  i : Byte;
  w : Word;
  id : String[17];
  vdata : Array [0..22] Of Byte;
  pack : packet;
  temp : ppacket;
Label
  failure;
Begin
  With x^ Do
  Begin
    id :=Short (dir)+'\'+Short (name);
    Status (24,'Get "'+id+'" ('+NumberHi (6,size)+' bytes)');
    vdata[0] :=0;
    vdata[1] :=0;
    vdata[2] :=0;
    vdata[3] :=0;
    vdata[4] :=typ;
    vdata[5] :=Length (id);
    For i :=1 To Length (id) Do
      vdata[5+i] :=Ord (id[i]);
    pack.typ :=$A2;
    pack.data :=@vdata;
    pack.size :=6+Length (id);

    If Not Skip_TI_reset Then
    Begin
      TIsend_command (c_request,$00,$00);
      If TIerror>0 Then Goto failure;
      If (TIget_command<>$56) And (TIerror=0) Then
        TIerror :=5;
      If TIerror>0 Then Goto failure;
    End;

    TIsend_packet (@pack,False);
    If TIerror>0 Then Goto failure;

    If (TIget_command<>$56) And (TIerror=0) Then
      TIerror :=5;
    If TIerror>0 Then Goto failure;

    TIget_packet (@pack,False,True);
    If TIerror>0 Then Goto failure;
    FreeMem (pack.data,pack.size); Inc (_mfree,pack.size);

    TIsend_command (c_OK,$00,$00);
    If TIerror>0 Then Goto failure;

    TIsend_command (c_wait,$00,$00);
    If TIerror>0 Then Goto failure;

    If (TIget_command<>$56) And (TIerror=0) Then
      TIerror :=5;
    If TIerror>0 Then Goto failure;

    TIget_packet (@pack,True,True);
    If TIerror>0 Then Goto failure;
    data :=pack.data;
    dsize :=pack.size;

    TIsend_command (c_OK,$00,$00);
    If TIerror>0 Then Goto failure;

    If (TIget_command<>$92) And (TIerror=0) Then
      TIerror :=5;
    If TIerror>0 Then Goto failure;

    TIsend_command (c_OK,$00,$00);
    If TIerror>0 Then Goto failure;
  End;
{  Status (24,'');}
  failure:
End;

Procedure TIsend_var (x : pTIvar);
Var
  i : Byte;
  id : String[17];
  vdata : Array [0..22] Of Byte;
  pack : packet;
Label
  failure;
Begin
  With x^ Do
  Begin
    id :=Short (dir)+'\'+Short (name);
    Status (24,'Send "'+id+'" ('+NumberHi (6,size)+' bytes)');
    vdata[0] :=size And $FF;
    vdata[1] :=(size Shr 8) And $FF;
    vdata[2] :=(size Shr 16) And $FF;
    vdata[3] :=(size Shr 24) And $FF;
    vdata[4] :=typ;
    vdata[5] :=Length (id);
    For i :=1 To Length (id) Do
      vdata[5+i] :=Ord (id[i]);
    pack.typ :=$06;
    pack.data :=@vdata;
    pack.size :=6+Length (id);

    TIsend_packet (@pack,False);
    If TIerror>0 Then Goto failure;

    If (TIget_command<>$56) And (TIerror=0) Then
      TIerror :=5;
    If TIerror>0 Then Goto failure;

    If (TIget_command<>$09) And (TIerror=0) Then
      TIerror :=5;
    If TIerror>0 Then Goto failure;

    TIsend_command (c_OK,$00,$00);
    If TIerror>0 Then Goto failure;

    pack.typ :=$15;
    pack.size :=dsize;
    pack.data :=data;
    TIsend_packet (@pack,True);
    If TIerror>0 Then Goto failure;

    If (TIget_command<>$56) And (TIerror=0) Then
      TIerror :=5;
    If TIerror>0 Then Goto failure;

    TIsend_command (c_eot,$00,$00);
    If TIerror>0 Then Goto failure;

    If (TIget_command<>$56) And (TIerror=0) Then
      TIerror :=5;
    If TIerror>0 Then Goto failure;
  End;
  Status (24,'');
  failure:
End;

Function TIget_backup (Var p : mpack; Var ROM : TIstr) : Word;
Var
  i : Byte;
  x,w : Word;
  _size,sum : LongInt;
  ver : String[8];
  pack : packet;
Label
  failure;
Begin
  TIreset;
  Dialog (26,11,53,13,' Retrieving backup ');
  TIsend_packet (@p_bckp_request,False);
  If TIerror>0 Then Goto failure;
  If (TIget_command<>$56) And (TIerror=0) Then
    TIerror :=5;
  If TIerror>0 Then Goto failure;

  x :=0;
  sum :=0;
  Repeat
    TIget_packet (@pack,False,True);
    If TIerror>0 Then Goto failure;
    With pack Do
    Begin
      _size :=data^[0];
      Inc (_size,data^[1] Shl 8);
      Inc (_size,data^[2] Shl 16);
      Inc (_size,data^[3] Shl 24);
      If (x=0) Then
      Begin
        For i :=1 To data^[5] Do
          ver[i] :=Chr (data^[5+i]);
        ver[0] :=Chr (data^[5]);
        Text (27,12,'ROM '+ver);
        For i :=1 To Length (ver) Do
          ROM[i] :=ver[i];
        If Length (ver)<8 Then
          ROM[Length (ver)+1] :=#0;
      End;
    End;
    FreeMem (pack.data,pack.size); Inc (_mfree,pack.size);

    TIsend_command (c_OK,$00,$00);
    If TIerror>0 Then Goto failure;
    If (_size=0) Then Goto failure;

    TIsend_command (c_wait,$00,$04);
    If TIerror>0 Then Goto failure;
    If (TIget_command<>$56) And (TIerror=0) Then
      TIerror :=5;
    If TIerror>0 Then Goto failure;

    TIget_packet (@pack,False,True);
    If TIerror>0 Then Goto failure;
    Inc (x);
    _size :=pack.size;
    With p[x] Do
    Begin
      size :=_size;
      data :=pack.data;
    End;
    Inc (sum,_size);
    Text (40,12,NumberHi (6,sum)+' bytes');

    TIsend_command (c_OK,$00,$00);
    If TIerror>0 Then Goto failure;
  Until (_size<1024+4);

  If (TIget_command<>$92) And (TIerror=0) Then
    TIerror :=5;
  If TIerror>0 Then Goto failure;
  TIsend_command (c_OK,$00,$00);
  If TIerror>0 Then Goto failure;

  failure:
  TIreset;
  End_dialog;
  TIget_backup :=x;
End;

Procedure TIsend_backup (fname : String);
Var
  f : File;
  nb,x,w : Word;
  _size,l : LongInt;
  hback : Record
            size : LongInt;
            typ  : Byte;
          End;
  head92b : h92b;
  p : mpack;
  s : String[16];
Const
  hpack : packet = (typ: 6; size:5);
Label
  failure;
Begin
  nb :=0;
  Dialog (28,10,51,14,' Sending backup ');
  Assign (f,fname);
  Reset (f,1);
  BlockRead (f,head92b,SizeOf (h92b));
  With head92b Do
  Begin
    Seek (f,ptr);
    _size :=fsize-ptr-2;
    Text (36,11,'ROM '+PascalStr (ROM,8));
  End;
  l :=_size;
  Repeat
    If l>=1024 Then
      w :=1024
    Else
      w :=l;
    Inc (nb);
    With p[nb] Do
    Begin
      size :=w{+4};
      GetMem (data,w); Inc (_malloc,w);
      BlockRead (f,data^{[4]},w);
{      data^[0] :=0;
      data^[1] :=4;
      data^[2] :=0;
      data^[3] :=0;}
      typ :=$15;
    End;
    Dec (l,1024);
  Until (l<=0);

  With hback Do
  Begin
    size :=_size;
    typ :=$1D;
  End;
  hpack.data :=@hback;
  TIsend_packet (@hpack,False);
  If TIerror>0 Then Goto failure;
  If (TIget_command<>$56) And (TIerror=0) Then
    TIerror :=5;
  If TIerror>0 Then Goto failure;

  hback.size :=0;
  w :=0;
  l :=0;
  s :=Number (6,_size)+' bytes';
  Repeat
    TIsend_packet (@hpack,False);
    If TIerror>0 Then Goto failure;

    If (TIget_command<>$56) And (TIerror=0) Then
      TIerror :=5;
    If TIerror>0 Then Goto failure;
    If (TIget_command<>$09) And (TIerror=0) Then
      TIerror :=5;
    If TIerror>0 Then Goto failure;
    TIsend_command (c_OK,$00,$00);

    Inc (w);
    Inc (l,p[w].size{-4});
    TIsend_packet (@p[w],False);
    Text (29,12,Number (6,l)+' / '+s);
    Text (37,13,Number (3,Round (100*l/_size))+'%');

    If (TIget_command<>$56) And (TIerror=0) Then
      TIerror :=5;
    If TIerror>0 Then Goto failure;
  Until (w>=nb);
  TIsend_command (c_eot,$00,$00);

  failure:
  Close (f);
  End_dialog;
  For w :=1 To nb Do
    With p[w] Do Begin
      FreeMem (data,size); Inc (_mfree,size); End;
End;

End.
